from dataclasses import dataclass
import autograd.numpy as np
from autograd import jacobian, hessian
from autograd.numpy import sin, cos, arctan, pi, arctan2

class SinCos:

    dt : float = 1.0

    def __init__(self, state_outlier_flag=False, 
                measurement_outlier_flag=False, noise_type='Gaussian'):

        self.dim_x = 2
        self.dim_y = 2
        self.P0 = 25 * np.eye(self.dim_x)
        self.x0 = np.random.multivariate_normal(mean=np.zeros(self.dim_x), cov=self.P0)
        
        self.state_outlier_flag = state_outlier_flag
        self.measurement_outlier_flag = measurement_outlier_flag
        self.noise_type = noise_type
        self.Q = np.eye(self.dim_x)
        self.R = np.eye(self.dim_y)
        

    def f(self, x):
        #  parmater
        k1 = 0.1
        k2 = 0.1
        F = np.eye(self.dim_x) + k1 * np.array([[-1,0],[0.1,-1]])
        x_ = F @ x + k2 * cos(x)
        return x_

    def h(self, x):
        y = x + sin(x)
        return y
    
    def f_withnoise(self, x):
        if self.state_outlier_flag:
            prob = np.random.rand()
            if prob <= 0.95:
                cov = self.Q  # 95%概率使用Q
            else:
                cov = 100 * self.Q  # 5%概率使用100Q
        else:
            cov = self.Q
        return self.f(x) + np.random.multivariate_normal(mean=np.zeros(self.dim_x), cov=cov)
    
    def h_withnoise(self, x):
        if self.noise_type == 'Gaussian':
            if self.measurement_outlier_flag:
                prob = np.random.rand()
                if prob <= 0.9:
                    cov = self.R  # 95%概率使用R
                else:
                    cov = 100 * self.R  # 5%概率使用100R
            else:
                cov = self.R
            return self.h(x) + np.random.multivariate_normal(mean=np.zeros(self.dim_y), cov=cov)
        else:
            return self.h(x) + np.random.laplace(loc=0, scale=1, size=(self.dim_y, ))

    def jac_f(self, x_hat):
        return jacobian(lambda x: self.f(x))(x_hat)
    
    def jac_h(self, x_hat):
        return jacobian(lambda x: self.h(x))(x_hat)


from filterpy.kalman import ExtendedKalmanFilter
import numpy as np
from scipy.linalg import sqrtm

class HuberEKF(ExtendedKalmanFilter):
    def __init__(self, model):
        super().__init__(dim_x=model.dim_x,
                         dim_z=model.dim_y,
                         )

        self.f = model.f
        self.h = model.h
        self.jac_f = model.jac_f
        self.jac_h = model.jac_h

        self.Q = model.Q
        self.R = model.R

        self.x = model.x0
        self.P = model.P0 

    def predict(self, u=0):
        F = self.jac_f(self.x)
        self.x = self.f(self.x)
        self.P = F @ self.P @ F.T + self.Q
        
        self.x_prior = self.x.copy()
        self.P_prior = self.P.copy()

    def _rho_function(self, x, r=1):
        for i in range(len(x)):
            if abs(x[i]) < r:
                x[i] = 0.5 * x[i] ** 2
            else:
                x[i] = r * abs(x[i]) - 0.5 * r * r
        return x
    
    def _phi_function(self, x, r=1):
        for i in range(len(x)):
            if abs(x[i]) < r:
                x[i] = 1
            else:
                x[i] = r * np.sign(x[i]) / x[i] 
        return x
    
    def update(self, y, max_iter=10, epson=1e-2):
        H = self.jac_h(self.x)
        x = self.x.copy()
        n, m = self.dim_z, self.dim_x
        zero_n_m = np.zeros((n, m))
        zero_m_n = np.zeros((m, n))
        SI = np.block([[self.R, zero_n_m],
                    [zero_m_n, self.P]])
        SI_inv_sqrt = np.linalg.inv(np.real(sqrtm(SI)))
        M = SI_inv_sqrt @ np.block([[H], [np.eye(self.dim_x)]])
        P_sqrt = SI_inv_sqrt[self.R.shape[0]:, self.R.shape[0]:]
        R_sqrt = SI_inv_sqrt[:self.R.shape[0], :self.R.shape[0]]

        x_ = 100 * np.ones(self.dim_x)
        
        for i in range(max_iter):
            if(np.sum(abs(x_ - x)) < epson):
                break
            x_ = x
            z = SI_inv_sqrt @ np.block([y, x]) 
            # x = np.linalg.inv(M.T @ M) @ M.T   
            e = z - M @ x
            phi = self._phi_function(e)
            Phi = np.diag(phi)
            hx = self.h(x)
            P_ = P_sqrt @ np.linalg.inv(Phi[self.R.shape[0]:, self.R.shape[0]:]) @ P_sqrt
            R_ = R_sqrt @ np.linalg.inv(Phi[:self.R.shape[0], :self.R.shape[0]]) @ R_sqrt
            PHT = P_ @ H.T
            self.S = H @ PHT + R_
            self.K = PHT @ np.linalg.inv(self.S)
            x = x + self.K @ (y - hx)
            # print(i, x)
        self.x = self.x + self.K @ (y - hx)
        I_KH = self._I - self.K @ H
        self.P = (I_KH @ P_ @ I_KH.T) + (self.K @ self.R @ self.K.T)

        self.x_post = self.x.copy()
        self.P_post = self.P.copy()


import os
from dataclasses import dataclass
from dataclasses_json import dataclass_json
from itertools import product
from scipy.linalg import expm
from scipy.linalg import LinAlgError, bandwidth
import autograd.numpy as np
from autograd import jacobian, hessian
from autograd.numpy import sin, cos, arctan, sqrt
# from base import Model
np.random.seed(42)


@dataclass_json
@dataclass
class Lorenz:
    
    dt: float = 0.02

    def __init__(self, state_outlier_flag=False, 
                measurement_outlier_flag=False, noise_type='Gaussian'):
        self.dim_x = 3
        self.dim_y = 5
        self.P0 = 0.01 * np.eye(self.dim_x)
        self.x0 =  np.random.multivariate_normal(mean=np.zeros(self.dim_x), cov=self.P0)

        self.state_outlier_flag = state_outlier_flag
        self.measurement_outlier_flag = measurement_outlier_flag
        self.noise_type = noise_type
        
        self.var = np.array([1e-2, 1e-2, 1e-2])
        self.obs_var = np.array([1e-2, 1e-4, 1e-4, 1e-4, 1e-4])
        self.Q = np.diag(self.var)
        self.R = np.diag(self.obs_var)

    def f(self, x):
        x_1 = x[0]
        A_x = np.array([
            [-10, 10, 0],
            [28, -1, -x_1],
            [0, x_1, -8/3]
        ])
        return expm(A_x * self.dt) @ x
    
    def h(self, x):
        rho = np.sqrt(np.sum(np.square(x)))
        r = np.sqrt(np.sum(np.square(x[:2])))
        cos_theta = x[0] / r
        sin_theta = x[1] / r
        cos_phi = r / rho
        sin_phi = x[2] / rho
        return np.array([rho, cos_theta, sin_theta, cos_phi, sin_phi])
    
    def f_withnoise(self, x):
        return self.f(x) + np.random.multivariate_normal(mean=np.zeros(self.dim_x), cov=self.Q)
    
    def h_withnoise(self, x):
        return self.h(x) + np.random.multivariate_normal(mean=np.zeros(self.dim_y), cov=self.R)
    
    def jac_f(self, x_hat):
        return jacobian(lambda x: self.f(x))(x_hat)
    
    def jac_h(self, x_hat):
        return jacobian(lambda x: self.h(x))(x_hat)


def expm(A):
    """Compute the matrix exponential of an array.

    Parameters
    ----------
    A : ndarray
        Input with last two dimensions are square ``(..., n, n)``.

    Returns
    -------
    eA : ndarray
        The resulting matrix exponential with the same shape of ``A``

    Notes
    -----
    Implements the algorithm given in [1], which is essentially a Pade
    approximation with a variable order that is decided based on the array
    data.

    For input with size ``n``, the memory usage is in the worst case in the
    order of ``8*(n**2)``. If the input data is not of single and double
    precision of real and complex dtypes, it is copied to a new array.

    For cases ``n >= 400``, the exact 1-norm computation cost, breaks even with
    1-norm estimation and from that point on the estimation scheme given in
    [2] is used to decide on the approximation order.

    References
    ----------
    .. [1] Awad H. Al-Mohy and Nicholas J. Higham, (2009), "A New Scaling
           and Squaring Algorithm for the Matrix Exponential", SIAM J. Matrix
           Anal. Appl. 31(3):970-989, :doi:`10.1137/09074721X`

    .. [2] Nicholas J. Higham and Francoise Tisseur (2000), "A Block Algorithm
           for Matrix 1-Norm Estimation, with an Application to 1-Norm
           Pseudospectra." SIAM J. Matrix Anal. Appl. 21(4):1185-1201,
           :doi:`10.1137/S0895479899356080`

    Examples
    --------
    >>> import numpy as np
    >>> from scipy.linalg import expm, sinm, cosm

    Matrix version of the formula exp(0) = 1:

    >>> expm(np.zeros((3, 2, 2)))
    array([[[1., 0.],
            [0., 1.]],
    <BLANKLINE>
           [[1., 0.],
            [0., 1.]],
    <BLANKLINE>
           [[1., 0.],
            [0., 1.]]])

    Euler's identity (exp(i*theta) = cos(theta) + i*sin(theta))
    applied to a matrix:

    >>> a = np.array([[1.0, 2.0], [-1.0, 3.0]])
    >>> expm(1j*a)
    array([[ 0.42645930+1.89217551j, -2.13721484-0.97811252j],
           [ 1.06860742+0.48905626j, -1.71075555+0.91406299j]])
    >>> cosm(a) + 1j*sinm(a)
    array([[ 0.42645930+1.89217551j, -2.13721484-0.97811252j],
           [ 1.06860742+0.48905626j, -1.71075555+0.91406299j]])

    """
    a = np.asarray(A)
    n = a.shape[-1]
    # Empty array
    if min(*a.shape) == 0:
        return np.empty_like(a)

    # Scalar case
    if a.shape[-2:] == (1, 1):
        return np.exp(a)

    if not np.issubdtype(a.dtype, np.inexact):
        a = a.astype(float)
    elif a.dtype == np.float16:
        a = a.astype(np.float32)

    # Explicit formula for 2x2 case, formula (2.2) in [1]
    # without Kahan's method numerical instabilities can occur.
    if a.shape[-2:] == (2, 2):
        a1, a2, a3, a4 = (a[..., [0], [0]],
                          a[..., [0], [1]],
                          a[..., [1], [0]],
                          a[..., [1], [1]])
        mu = sqrt((a1-a4)**2 + 4*a2*a3)/2.  # csqrt slow but handles neg.vals

        eApD2 = np.exp((a1+a4)/2.)
        AmD2 = (a1 - a4)/2.
        coshMu = np.cosh(mu)
        sinchMu = np.ones_like(coshMu)
        mask = mu != 0
        sinchMu[mask] = np.sinh(mu[mask]) / mu[mask]
        eA = np.empty((a.shape), dtype=mu.dtype)
        eA[..., [0], [0]] = eApD2 * (coshMu + AmD2*sinchMu)
        eA[..., [0], [1]] = eApD2 * a2 * sinchMu
        eA[..., [1], [0]] = eApD2 * a3 * sinchMu
        eA[..., [1], [1]] = eApD2 * (coshMu - AmD2*sinchMu)
        if np.isrealobj(a):
            return eA.real
        return eA

    # larger problem with unspecified stacked dimensions.
    n = a.shape[-1]
    eA = np.empty(a.shape, dtype=a.dtype)
    # working memory to hold intermediate arrays
    Am = np.empty((5, n, n), dtype=a.dtype)

    # Main loop to go through the slices of an ndarray and passing to expm
    for ind in product(*[range(x) for x in a.shape[:-2]]):
        aw = a[ind]

        lu = bandwidth(aw)
        if not any(lu):  # a is diagonal?
            eA[ind] = np.diag(np.exp(np.diag(aw)))
            continue

        # Generic/triangular case; copy the slice into scratch and send.
        # Am will be mutated by pick_pade_structure
        Am[0, :, :] = aw
        m, s = pick_pade_structure(Am)

        if s != 0:  # scaling needed
            Am[:4] *= [[[2**(-s)]], [[4**(-s)]], [[16**(-s)]], [[64**(-s)]]]

        pade_UV_calc(Am, n, m)
        eAw = Am[0]

        if s != 0:  # squaring needed

            if (lu[1] == 0) or (lu[0] == 0):  # lower/upper triangular
                # This branch implements Code Fragment 2.1 of [1]

                diag_aw = np.diag(aw)
                # einsum returns a writable view
                np.einsum('ii->i', eAw)[:] = np.exp(diag_aw * 2**(-s))
                # super/sub diagonal
                sd = np.diag(aw, k=-1 if lu[1] == 0 else 1)

                for i in range(s-1, -1, -1):
                    eAw = eAw @ eAw

                    # diagonal
                    np.einsum('ii->i', eAw)[:] = np.exp(diag_aw * 2.**(-i))
                    exp_sd = _exp_sinch(diag_aw * (2.**(-i))) * (sd * 2**(-i))
                    if lu[1] == 0:  # lower
                        np.einsum('ii->i', eAw[1:, :-1])[:] = exp_sd
                    else:  # upper
                        np.einsum('ii->i', eAw[:-1, 1:])[:] = exp_sd

            else:  # generic
                for _ in range(s):
                    eAw = eAw @ eAw

        # Zero out the entries from np.empty in case of triangular input
        if (lu[0] == 0) or (lu[1] == 0):
            eA[ind] = np.triu(eAw) if lu[0] == 0 else np.tril(eAw)
        else:
            eA[ind] = eAw

    return eA


# def show_sys_img(sys: Lorenz, x0: jax.Array):
#     x = x0
#     states = [x]
#     for _ in range(2000):
#         x = sys.transition(x)
#         states.append(x)
#     states = jax.device_get(jnp.stack(states))
#     fig = plt.figure(figsize=(4, 4), tight_layout=True)
#     ax = fig.add_subplot(111, projection='3d')
#     ax.plot3D(states[:, 0], states[:, 1], states[:, 2])
#     plt.savefig('lorenz.png')


if __name__ == '__main__':
    from matplotlib import pyplot as plt
    os.environ['XLA_PYTHON_CLIENT_PREALLOCATE'] = 'false'
    lorenz = Lorenz()
    # show_sys_img(lorenz, np.random.normal(size=(3,)))


import time

import autograd.numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
from autograd import hessian, jacobian
from filterpy.kalman import JulierSigmaPoints, MerweScaledSigmaPoints
from filterpy.kalman import KalmanFilter as KF
from filterpy.kalman import UnscentedKalmanFilter as UKF
from filterpy.kalman import unscented_transform as UT
from scipy.optimize import minimize

from .utils import (cal_mean, cal_mean_mc, is_positive_semidefinite,
                    kl_divergence)


class NANO:

    lr : float = 1.0 # 0.2
    threshold: float = 1e-4

    def __init__(self, model, loss_type='log_likelihood_loss', init_type='prior',
                 derivate_type='stein', iekf_max_iter=1, n_iterations=10, 
                    delta=5, c=5, beta=1e-4):    
        self.model = model
        self.dim_x = model.dim_x
        self.dim_y = model.dim_y    
        self.x = model.x0
        self.P = model.P0

        self.f = model.f
        self.h = model.h
        self.jac_f = model.jac_f
        self.jac_h = model.jac_h
        self.Q = model.Q
        self.R = model.R
        self._I = np.eye(self.dim_x)

        self.n_iterations = n_iterations
        self.points = MerweScaledSigmaPoints(self.dim_x, alpha=0.1, beta=2.0, kappa=1.0)
        self.x_prior = self.x
        self.P_prior = self.P
        self.x_post = self.x
        self.P_post = self.P

        self.init_type = init_type
        self.derivate_type = derivate_type
        self.iekf_max_iter = iekf_max_iter
        self.delta = delta
        self.c = c
        self.beta = beta

        if loss_type == 'pseudo_huber_loss':
            self.loss_func = self.pseudo_huber_loss
        elif loss_type == 'weighted_log_likelihood_loss':
            self.loss_func = self.weighted_log_likelihood_loss
        elif loss_type == 'beta_likelihood_loss':
            self.loss_func = self.beta_likelihood_loss
        else:
            self.loss_func = self.log_likelihood_loss
        
    def log_likelihood_loss(self, x, y):
        return 0.5 * np.dot(y - self.h(x), np.dot(np.linalg.inv(self.R), y - self.h(x)))
    
    def pseudo_huber_loss(self, x, y):
        delta = self.delta
        mse_loss = np.dot(y - self.h(x), np.dot(np.linalg.inv(self.R), y - self.h(x)))
        return delta**2 * (np.sqrt(1 + mse_loss / delta**2) - 1)
    
    def weighted_log_likelihood_loss(self, x, y):
        c = self.c
        mse_loss = 0.5 * np.dot(y - self.h(x), np.dot(np.linalg.inv(self.R), y - self.h(x)))
        weight = 1/(1 + mse_loss / c**2)
        return weight * mse_loss
    
    def beta_likelihood_loss(self, x, y):
        beta = self.beta
        R_inv = np.linalg.inv(self.R)
        det_R = np.linalg.det(self.R)
        return 1 / ((beta + 1)**1.5*(2*np.pi)**(self.dim_y*beta/2)) * det_R**(beta / 2) \
                - (1 / beta + 1) / ((2 * np.pi)**(beta*self.dim_y/2) * det_R**(beta/2)) * np.exp(-0.5*beta*(y-self.h(x)).T @ R_inv @ (y-self.h(x)))

    def loss_func_jacobian(self, x, y):
        # cal jacobian of loss function
        return jacobian(lambda x: self.loss_func(x, y))(x)
        
    def loss_func_hessian(self, x, y):
        # cal hessian of loss function
        return hessian(lambda x: self.loss_func(x, y))(x)
    
    def loss_func_hessian_diff(self, x, y, epsilon=5e-5):
        n = len(x)
        Hessian = np.zeros((n, n))
        f = self.loss_func
        fx = f(x, y)
        
        for i in range(n):
            for j in range(i, n):
                x_ij = x.copy()
                x_ij[i] += epsilon
                x_ij[j] += epsilon
                fij = f(x_ij, y)
                
                x_i = x.copy()
                x_i[i] += epsilon
                fi = f(x_i, y)
                
                x_j = x.copy()
                x_j[j] += epsilon
                fj = f(x_j, y)
                
                Hessian[i, j] = (fij - fi - fj + fx) / (epsilon**2)
                Hessian[j, i] = Hessian[i, j]
                
        return Hessian
    
    def map_loss(self, x_prior, P_prior, x_posterior, y):
        l1 = 0.5 * (x_posterior - x_prior).T @ np.linalg.inv(P_prior) @ (x_posterior - x_prior) 
        l2 = self.loss_func(x_posterior, y)
        return l1 + l2
            
    def update_init(self, y, x_prior, P_prior):
        # Laplace Approximation
        loss = lambda x_posterior: self.map_loss(x_prior, P_prior, x_posterior, y)
        x_hat_posterior = minimize(loss, x0=x_prior, method='BFGS').x
        P_posterior_inv = hessian(lambda x: self.map_loss(x_prior, P_prior, x, y))(x_hat_posterior)
        return x_hat_posterior, P_posterior_inv
    
    def update_iekf_init(self, y, x_prior, P_prior, max_iter=1):
        # Iterated Extended Kalman Filter (IEKF) for Maximum A Posteriori (MAP)
        x_hat = x_prior
        for i in range(max_iter):
            H = self.jac_h(x_hat)
            hx = self.h(x_hat)
            v = y - hx - H @ (x_prior - x_hat)
            PHT = P_prior @ H.T
            S = H @ PHT + self.R
            K = PHT @ np.linalg.inv(S)
            x_hat = x_prior + K @ v
        
        x_hat_posterior = x_hat
        I_KH = self._I - K @ H
        P_posterior = (I_KH @ P_prior @ I_KH.T) + (K @ self.R @ K.T)
        P_posterior_inv = np.linalg.inv(P_posterior)
        
        return x_hat_posterior, P_posterior_inv
    
    def predict(self, u=0):
        sigmas = self.points.sigma_points(self.x, self.P)

        self.sigmas_f = np.zeros((len(sigmas), self.dim_x))
        for i, s in enumerate(sigmas):
            self.sigmas_f[i] = self.f(s, u)        
        
        self.x, self.P = UT(self.sigmas_f, self.points.Wm, self.points.Wc, self.Q)

        is_positive_semidefinite(self.P)

        self.x_prior = self.x.copy()
        self.P_prior = self.P.copy()
    
    def update(self, y, lr = None, n_iterations = None):
        if lr == None:
            lr = self.lr
        if n_iterations is None:
            n_iterations = self.n_iterations
        
        # Initialize the first iteration step
        x_hat_prior = self.x.copy()
        P_inv_prior = np.linalg.inv(self.P).copy()
        if self.init_type == 'prior':
            x_hat, P_inv = x_hat_prior, P_inv_prior
        elif self.init_type == 'laplace':
            x_hat, P_inv = self.update_init(y, x_hat_prior, self.P.copy())
        else: # self.init_type == 'iekf'
            x_hat, P_inv = self.update_iekf_init(y, x_hat_prior, self.P.copy(), self.iekf_max_iter)
        
        is_positive_semidefinite(P_inv)

        for _ in range(n_iterations):          
            P = np.linalg.inv(P_inv)
            is_positive_semidefinite(P)

            if self.derivate_type == 'stein':
                E_hessian = P_inv @ cal_mean(lambda x: np.outer(x-x_hat, x-x_hat)*self.loss_func(x,y), x_hat, P, self.points) @ P_inv \
                            - cal_mean(lambda x: self.loss_func(x, y), x_hat, P, self.points) * P_inv
                P_inv_next = P_inv_prior - lr * E_hessian
                is_positive_semidefinite(P_inv_next)
                P_next = np.linalg.inv(P_inv_next)
                x_hat_next = x_hat - lr*(P_next @ P_inv @ cal_mean(lambda x: (x - x_hat) * self.loss_func(x, y), x_hat, P, self.points) - P_next @ P_inv_prior @ (x_hat - x_hat_prior))

            else: # self.derivate_type == 'direct'
                P_inv_next = P_inv_prior + lr*cal_mean(lambda x: self.loss_func_hessian_diff(x, y), x_hat, P, self.points)
                is_positive_semidefinite(P_inv_next)
                P_next = np.linalg.inv(P_inv_next)
                x_hat_next = x_hat - lr*(P_next @ cal_mean(lambda x: self.loss_func_jacobian(x, y), x_hat, P, self.points) - P_next @ P_inv_prior @ (x_hat - x_hat_prior))

            kld = kl_divergence(x_hat, P, x_hat_next, P_next)
            if kld < self.threshold:
                P_inv = P_inv_next.copy()
                x_hat = x_hat_next.copy()
                break

            P_inv = P_inv_next.copy()
            x_hat = x_hat_next.copy()
            
        self.x = x_hat
        self.P = np.linalg.inv(P_inv)

        self.x_post = self.x.copy()
        self.P_post = self.P.copy()



from filterpy.kalman import ExtendedKalmanFilter
import numpy as np



class IEKF(ExtendedKalmanFilter):
    def __init__(self, model, max_iter=3):
        super().__init__(dim_x=model.dim_x,
                         dim_z=model.dim_y,
                         )

        self.f = model.f
        self.h = model.h
        self.jac_f = model.jac_f
        self.jac_h = model.jac_h
        self.max_iter = max_iter

        self.Q = model.Q
        self.R = model.R

        self.x = model.x0
        self.P = model.P0 

    def predict(self, u=0):
        F = self.jac_f(self.x, u)
        self.x = self.f(self.x, u)
        self.P = F @ self.P @ F.T + self.Q
        
        self.x_prior = self.x.copy()
        self.P_prior = self.P.copy()
    
    def update(self, y, max_iter=None):
        if max_iter == None:
            max_iter = self.max_iter

        x_hat = self.x
        P_prior = self.P_prior
        for i in range(max_iter):
            H = self.jac_h(self.x)
            hx = self.h(self.x)
            v = y - hx - H @ (x_hat - self.x)
            PHT = P_prior @ H.T
            self.S = H @ PHT + self.R
            self.K = PHT @ np.linalg.inv(self.S)
            # x_r= self.K @ (y - hx - H @ (x_hat - self.x))
            self.x = x_hat + self.K @ v
        I_KH = self._I - self.K @ H
        # self.P = P_prior - self.K @ self.S @ self.K.T
        self.P = (I_KH @ P_prior @ I_KH.T) + (self.K @ self.R @ self.K.T)
        
        self.x_post = self.x.copy()
        self.P_post = self.P.copy()

    # def update(self, y, max_iter=2):
    #     x_hat = self.x
    #     for i in range(max_iter):
    #         H = self.jac_h(self.x)
    #         hx = self.h(self.x)
    #         PHT = self.P @ H.T
    #         self.S = H @ PHT + self.R
    #         self.K = PHT @ np.linalg.inv(self.S)
    #         x_r= self.K @ (y - hx - H @ (x_hat - self.x))
    #         self.x = x_hat + x_r
    #         I_KH = self._I - self.K @ H
    #         self.P = (I_KH @ self.P @ I_KH.T) + (self.K @ self.R @ self.K.T)
        
    #     self.x_post = self.x.copy()
    #     self.P_post = self.P.copy()


from dataclasses import dataclass
import autograd.numpy as np
from autograd import jacobian, hessian
from autograd.numpy import sin, cos, arctan, pi, arctan2, sqrt

class BlackHole:
    
    dt : float = 0.02

    def __init__(self, state_outlier_flag=False, 
                measurement_outlier_flag=False, noise_type='Gaussian'):
        # System size
        self.dim_x = 2
        self.dim_y = 1
        self.P0 = np.eye(self.dim_x)
        self.x0 = np.random.multivariate_normal(mean=np.zeros(self.dim_x), cov=self.P0)
        

        self.state_outlier_flag = state_outlier_flag
        self.measurement_outlier_flag = measurement_outlier_flag
        self.noise_type = noise_type
        self.Q = np.diag(np.array([1e-2, 1e-2]))
        self.R = np.diag(np.array([1e-2]))
        
        self.I_last = None

        # Other parameters
        self.M = 1.0
        self.e = 0.5
        self.p = 100
    
    def f(self, x):
        """Transition function for the model."""
        x_ = rk4(self.dxdt, x, 0, self.dt)
        return x_

    def h(self, x):
        """Observation function for the model."""
        # ddI = second_order_fd(self.get_quadrupole(states, 0), self.dt)
        I = self.get_quadrupole(x)
        # if I_last == None:
        state_next1 = self.f(x)
        I_next1 = self.get_quadrupole(state_next1)
        state_next2 = self.f(state_next1)
        I_next2 = self.get_quadrupole(state_next2)
        state_next3 = self.f(state_next2)
        I_next3 = self.get_quadrupole(state_next3)
        ddI = (2 * I - 5 * I_next1 + 4 * I_next2 - I_next3) / self.dt ** 2
        # else:
        #     I_next = self.get_quadrupole(self.transition(states))
        #     ddI = (I_next - 2 * I + I_last) / self.dt ** 2
        # self.I_last = I
        # print(ddI.shape)
        ddIxx, ddIyy, ddIxy = ddI
        return np.array([(ddIxx - ddIyy) * np.sqrt(4 * np.pi / 5)])
    
    def f_withnoise(self, x):
        if self.state_outlier_flag:
            prob = np.random.rand()
            if prob <= 0.95:
                cov = self.Q  # 95%概率使用Q
            else:
                cov = 100 * self.Q  # 5%概率使用100Q
        else:
            cov = self.Q
        return self.f(x) + np.random.multivariate_normal(mean=np.zeros(self.dim_x), cov=cov)
    
    def h_withnoise(self, x):
        if self.noise_type == 'Gaussian':
            if self.measurement_outlier_flag:
                prob = np.random.rand()
                if prob <= 0.9:
                    cov = self.R  # 95%概率使用R
                else:
                    cov = 100 * self.R  # 5%概率使用100R
            else:
                cov = self.R
            return self.h(x) + np.random.multivariate_normal(mean=np.zeros(self.dim_y), cov=cov)
        else:
            return self.h(x) + np.random.laplace(loc=0, scale=1, size=(self.dim_y, ))

    def jac_f(self, x_hat):
        return jacobian(lambda x: self.f(x))(x_hat)
    
    def jac_h(self, x_hat):
        return jacobian(lambda x: self.h(x))(x_hat)
    
    def dxdt(self, x, t):
        M, e, p = self.M, self.e, self.p
        phi, chi = x
        dphi = (p - 2 - 2 * e * cos(chi)) * (1 + e * cos(chi)) ** 2
        dphi /= M * p ** (3 / 2) * sqrt((p - 2) ** 2 - 4 * e**2)
        dchi = (
            (p - 2 - 2 * e * cos(chi))
            * (1 + e * cos(chi)) ** 2
            * sqrt(p - 6 - 2 * e * cos(chi))
        )
        dchi /= M * p**2 * sqrt((p - 2) ** 2 - 4 * e**2)
        return np.array([dphi, dchi])

    def euclidean_norm(self, chi):
        M, e, p = self.M, self.e, self.p
        r = p * M / (1 + e * cos(chi))
        return r
    
    def convert_to_trajectories(self, x):
        phi, chi = x
        r = self.euclidean_norm(chi)
        r_2 = np.array([r * cos(phi), r * sin(phi)])
        return r_2

    def get_quadrupole(self, x):
        r = self.convert_to_trajectories(x)
        x, y = r
        Ixx = x**2 * self.M
        Iyy = y**2 * self.M
        Ixy = x * y * self.M
        return np.array([Ixx, Iyy, Ixy])

def rk4(f, x, t, dt, stages=4, s=0):
    """Runge-Kutta (explicit, non-adaptive) numerical (S)ODE solvers.

    For ODEs, the order of convergence equals the number of `stages`.

    For SDEs with additive noise (`s>0`), the order of convergence
    (both weak and strong) is 1 for `stages` equal to 1 or 4.
    These correspond to the classic Euler-Maruyama scheme and the Runge-Kutta
    scheme for S-ODEs respectively, see `bib.grudzien2020numerical`
    for a DA-specific discussion on integration schemes and their discretization errors.

    Parameters
    ----------
    f : function
        The time derivative of the dynamical system. Must be of the form `f(t, x)`

    x : ndarray or float
        State vector of the forcing term

    t : float
        Starting time of the integration

    dt : float
        Integration time step.

    stages : int, optional
        The number of stages of the RK method.
        When `stages=1`, this becomes the Euler (-Maruyama) scheme.
        Default: 4.

    s : float
        The diffusion coeffient (std. dev) for models with additive noise.
        Default: 0, yielding deterministic integration.

    Returns
    -------
    ndarray
        State vector at the new time, `t+dt`
    """

    # Draw noise
    if s > 0:
        W = s * np.sqrt(dt) * np.random.randn(*x.shape)
    else:
        W = 0

    # Approximations to Delta x
    if stages >= 1: k1 = dt * f(x,           t)         + W    # noqa
    if stages >= 2: k2 = dt * f(x+k1/2.0,    t+dt/2.0)  + W    # noqa
    if stages == 3: k3 = dt * f(x+k2*2.0-k1, t+dt)      + W    # noqa
    if stages == 4:
                    k3 = dt * f(x+k2/2.0,    t+dt/2.0)  + W    # noqa
                    k4 = dt * f(x+k3,        t+dt)      + W    # noqa

    # Mix proxies
    if    stages == 1: y = x + k1                              # noqa
    elif  stages == 2: y = x + k2                              # noqa
    elif  stages == 3: y = x + (k1 + 4.0*k2 + k3)/6.0          # noqa
    elif  stages == 4: y = x + (k1 + 2.0*(k2 + k3) + k4)/6.0   # noqa
    else:
        raise NotImplementedError

    return y

# if __name__ == "__main__":
#     blackhole = BlackHole()
#     x0 = np.random.normal(size=(2, ))
#     print(blackhole.transition(x0), blackhole.observation(x0))



from .EKF import EKF
from .UKF import UKF
from .NANO import NANO
from .HuberEKF import HuberEKF
from .iEKF import IEKF
from .PF import PF
from .PLF import PLF
from .utils import is_positive_semidefinite, cal_mean


import autograd.numpy as np
from .model import Model


class WienerVelocity(Model):

    dt : float = 0.1
    def __init__(self, state_outlier_flag=False, 
                measurement_outlier_flag=False, noise_type='Gaussian'):
        super().__init__(self)
        self.F = np.array([[1, 0, self.dt, 0],
                           [0, 1, 0, self.dt],
                           [0, 0, 1, 0],
                           [0, 0, 0, 1],
                           ])
        self.H = np.array([[1, 0, 0, 0],
                           [0, 1, 0, 0]])

        self.dim_x = self.F.shape[0]
        self.dim_y = self.H.shape[0]
        self.x0 = np.array([0., 0., 1., 1.])
        self.P0 = np.eye(self.dim_x)

        self.state_outlier_flag = state_outlier_flag
        self.measurement_outlier_flag = measurement_outlier_flag
        self.noise_type = noise_type
        self.alpha = 2.0
        self.beta = 5.0

        dt = self.dt
        self.var = np.array([1, 1, 1, 1])
        self.obs_var = np.array([1, 1])
        self.Q = np.array([
            [dt**3/3, 0, dt**2/2, 0],
            [0, dt**3/3, 0, dt**2/2],
            [dt**2/2, 0, dt, 0],
            [0, dt**2/2, 0, dt]
        ])
        if noise_type == 'Beta':
            self.R = np.eye(self.dim_y) * (self.alpha * self.beta) / ((self.alpha + self.beta) ** 2 * (self.alpha + self.beta + 1))
        else:
            self.R = np.diag(self.obs_var)


    def f(self, x, dt=None, u=None):
        if u is None:
            return self.F @ x

    def h(self, x):
        return self.H @ x

    def jac_f(self, x, u=None):
        return self.F

    def jac_h(self, x):
        return self.H

    def f_withnoise(self, x):
        if self.state_outlier_flag:
            prob = np.random.rand()
            if prob <= 0.9:
                cov = self.Q  # 95%概率使用Q
            else:
                cov = 100 * self.Q  # 5%概率使用100Q
        else:
            cov = self.Q
        return self.f(x) + np.random.multivariate_normal(mean=np.zeros(self.dim_x), cov=cov)
    
    def h_withnoise(self, x):
        if self.noise_type == 'Gaussian':
            if self.measurement_outlier_flag:
                prob = np.random.rand()
                if prob <= 0.9:
                    cov = self.R  # 95%概率使用R
                else:
                    cov = 1000 * self.R  # 5%概率使用100R
            else:
                cov = self.R
            return self.h(x) + np.random.multivariate_normal(mean=np.zeros(self.dim_y), cov=cov)
        elif self.noise_type == 'Beta':
            noise = np.random.beta(self.alpha, self.beta, self.dim_y)
            noise = noise - np.mean(noise)
            return self.h(x) + noise
        else:
            return self.h(x) + np.random.laplace(loc=0, scale=1, size=(self.dim_y, ))


import autograd.numpy as np
from autograd import jacobian, hessian
from autograd.numpy import sin, cos, arctan, pi, arctan2
from .model import Model

class Air_Traffic(Model):

    q1 = 0.5
    q2 = 1e-6
    height = 50
    dt = 0.2

    def __init__(self, state_outlier_flag=False, 
                measurement_outlier_flag=False, noise_type='Gaussian'):
        super().__init__(self)
        q1 = self.q1
        q2 = self.q2
        height = self.height
        tau = self.dt
        
        self.dim_x = 5
        self.dim_y = 4
        self.P0 = np.diag(np.array([5, 5, 2e4, 10, 1e-7]))
        self.x0 =  np.array([130, 25, -20, 1, -4*pi/180])

        self.state_outlier_flag = state_outlier_flag
        self.measurement_outlier_flag = measurement_outlier_flag
        self.noise_type = noise_type
        self.alpha = 2.0
        self.beta = 5.0
        
        self.Q = np.array([
            [q1*tau**3/3, q1*tau**2/2, 0, 0, 0],
            [q1*tau**2/2, q1*tau, 0, 0, 0],
            [0, 0, q1*tau**3/3, q1*tau**2/2, 0],
            [0, 0, q1*tau**2/2, q1*tau, 0],
            [0, 0, 0, 0, q2*tau],
        ])
        self.obs_var = np.array([1000, (30*pi/180)**2, (30*pi/180)**2, 100])
        R1 = np.diag(np.array([1000, (30*pi/180)**2, (30*pi/180)**2, 100]))
        R2 = np.diag(np.array([1000, (1e-3*pi/180)**2, (30*pi/180)**2, 1e-4]))
        # self.R = R1
        if noise_type == 'Beta':
            self.R = np.eye(self.dim_y) * (self.alpha * self.beta) / ((self.alpha + self.beta) ** 2 * (self.alpha + self.beta + 1))
        else:
            self.R = R1

    def f(self, x, u=None):
        tau = self.dt
        Delta = x[4]
        F = np.array([
            [1, sin(Delta*tau)/Delta, 0, -(1 - cos(Delta*tau))/Delta, 0],
            [0, cos(Delta*tau), 0, -sin(Delta*tau), 0],
            [0, (1 - cos(Delta*tau))/Delta, 1, sin(Delta*tau)/Delta, 0],
            [0, sin(Delta*tau), 0, cos(Delta*tau), 0],
            [0, 0, 0, 0, 1]
        ])
        return F @ x
    
    def h(self, x):
        height = self.height
        px, dpx, py, dpy, Delta = x
        y1 = np.sqrt(px**2 + py ** 2 + height**2)
        y2 = arctan2(py, px)
        y3 = arctan(height / np.sqrt(px**2 + py ** 2))
        y4 = (px * dpx + py * dpy) / y1
        return np.array([y1, y2, y3, y4])
    
    def f_withnoise(self, x, u=None):
        if self.state_outlier_flag:
            prob = np.random.rand()
            if prob <= 0.95:
                cov = self.Q  # 95%概率使用Q
            else:
                cov = 100 * self.Q  # 5%概率使用100Q
        else:
            cov = self.Q
        return self.f(x, u) + np.random.multivariate_normal(mean=np.zeros(self.dim_x), cov=cov)
    
    def h_withnoise(self, x):
        if self.noise_type == 'Gaussian':
            if self.measurement_outlier_flag:
                prob = np.random.rand()
                if prob <= 0.9:
                    cov = self.R  # 95%概率使用R
                else:
                    cov = 100 * self.R  # 5%概率使用100R
            else:
                cov = self.R
            return self.h(x) + np.random.multivariate_normal(mean=np.zeros(self.dim_y), cov=cov)
        elif self.noise_type == 'Beta':
            noise = np.random.beta(self.alpha, self.beta, self.dim_y)
            noise = noise - np.mean(noise)
            return self.h(x) + noise
        else:
            return self.h(x) + np.random.laplace(loc=0, scale=self.obs_var, size=(self.dim_y, ))

    def jac_f(self, x_hat, u=0):
        return jacobian(lambda x: self.f(x))(x_hat)
    
    def jac_h(self, x_hat, u=0):
        return jacobian(lambda x: self.h(x))(x_hat)
        



import matplotlib.pyplot as plt
from filterpy.kalman import UnscentedKalmanFilter as UKF
from filterpy.kalman import MerweScaledSigmaPoints, JulierSigmaPoints
from filterpy.kalman import unscented_transform
import autograd.numpy as np
from autograd import jacobian, hessian
import scipy


def is_positive_definite(A):
    try:
        scipy.linalg.cholesky(A)
        return True
    except:
        return False


def f(x, dt):
    return [x[0] + x[2] * dt, x[1] + x[3] * dt, x[2], x[3]]


# 非线性观测函数
def h(x):
    r = np.sqrt(x[0] ** 2 + x[1] ** 2)
    theta = np.arctan2(x[1], x[0])
    return [r, theta]


def kl_divergence_gaussian(mu_p, Sigma_p, mu_q, Sigma_q):
    """
    Compute KL divergence of two multivariate Gaussian distributions.
    KL(p||q) for Gaussian p(mu_p, Sigma_p) and q(mu_q, Sigma_q)
    """
    k = len(mu_p)
    Sigma_q_inv = np.linalg.inv(Sigma_q)

    kl_div = 0.5 * (np.trace(Sigma_q_inv @ Sigma_p)
                    + (mu_q - mu_p).T @ Sigma_q_inv @ (mu_q - mu_p)
                    - k
                    + np.log(np.linalg.det(Sigma_q) / np.linalg.det(Sigma_p)))
    return kl_div


def plf_update(x_prior, P_prior, R, z):
    x_hat = x_prior
    P_hat = P_prior
    x_hat_old = 100 * np.ones_like(x_prior)
    P_hat_old = 100 * np.eye(4)
    ite = 0
    while kl_divergence_gaussian(x_hat_old, P_hat_old, x_hat, P_hat) >= 1e-2 and ite <= 100:
        ite += 1
        x_hat_old = x_hat
        P_hat_old = P_hat
        points_linear = MerweScaledSigmaPoints(n=4, alpha=1e-3, beta=2., kappa=0.)
        sigmas = points_linear.sigma_points(x_hat, P_hat)
        sigmas_h = np.empty((9, 2))
        for i in range(sigmas.shape[0]):
            sigmas_h[i] = h(sigmas[i])
        zp, Pz = unscented_transform(sigmas_h, points_linear.Wm, points_linear.Wc, noise_cov=None)
        Pxz = np.zeros((4, 2))
        for i in range(sigmas.shape[0]):
            Pxz += points_linear.Wc[i] * np.outer(sigmas[i] - x_hat, sigmas_h[i] - zp)

        H = Pxz.T @ np.linalg.inv(P_hat)
        b = zp - H @ x_hat
        Omega = Pz - H @ P_hat @ H.T

        x_hat = x_prior + P_prior @ H.T @ np.linalg.inv(H @ P_prior @ H.T + Omega + R) @ (z - H @ x_prior - b)
        P_hat = P_prior - P_prior @ H.T @ np.linalg.inv(H @ P_prior @ H.T + Omega + R) @ H @ P_prior

    return x_hat, P_hat


def plf_update(x_prior, P_prior, R, z):
    x_hat = x_prior
    P_hat = P_prior
    x_hat_old = 100 * np.ones_like(x_prior)
    P_hat_old = 100 * np.eye(4)
    ite = 0
    while kl_divergence_gaussian(x_hat_old, P_hat_old, x_hat, P_hat) >= 1e-2 and ite <= 100:
        ite += 1
        x_hat_old = x_hat
        P_hat_old = P_hat
        points_linear = MerweScaledSigmaPoints(n=4, alpha=1e-3, beta=2., kappa=0.)
        sigmas = points_linear.sigma_points(x_hat, P_hat)
        sigmas_h = np.empty((9, 2))
        for i in range(sigmas.shape[0]):
            sigmas_h[i] = h(sigmas[i])
        zp, Pz = unscented_transform(sigmas_h, points_linear.Wm, points_linear.Wc, noise_cov=None)
        Pxz = np.zeros((4, 2))
        for i in range(sigmas.shape[0]):
            Pxz += points_linear.Wc[i] * np.outer(sigmas[i] - x_hat, sigmas_h[i] - zp)

        H = Pxz.T @ np.linalg.inv(P_hat)
        b = zp - H @ x_hat
        Omega = Pz - H @ P_hat @ H.T

        x_hat = x_prior + P_prior @ H.T @ np.linalg.inv(H @ P_prior @ H.T + Omega + R) @ (z - H @ x_prior - b)
        P_hat = P_prior - P_prior @ H.T @ np.linalg.inv(H @ P_prior @ H.T + Omega + R) @ H @ P_prior

    return x_hat, P_hat


def rplf_update(x_prior, P_prior, R, z):
    x_hat = x_prior
    P_hat = P_prior
    x_hat_old = 100 * np.ones_like(x_prior)
    P_hat_old = 100 * np.eye(4)
    ite = 0
    alpha = 0.8
    while kl_divergence_gaussian(x_hat_old, P_hat_old, x_hat, P_hat) >= 1e-2 and ite <= 100:
        ite += 1
        x_hat_old = x_hat
        P_hat_old = P_hat
        points_linear = MerweScaledSigmaPoints(n=4, alpha=1e-3, beta=2., kappa=0.)
        sigmas = points_linear.sigma_points(x_hat, P_hat)
        sigmas_h = np.empty((9, 2))
        for i in range(sigmas.shape[0]):
            sigmas_h[i] = h(sigmas[i])
        zp, Pz = unscented_transform(sigmas_h, points_linear.Wm, points_linear.Wc, noise_cov=None)
        Pxz = np.zeros((4, 2))
        for i in range(sigmas.shape[0]):
            Pxz += points_linear.Wc[i] * np.outer(sigmas[i] - x_hat, sigmas_h[i] - zp)

        H = Pxz.T @ np.linalg.inv(P_hat)
        b = zp - H @ x_hat
        Omega = Pz - H @ P_hat @ H.T

        x_hat = x_prior + alpha*P_prior @ H.T @ np.linalg.inv(H @ P_prior @ H.T + Omega + R) @ (z - H @ x_prior - b)
        P_hat = P_prior - P_prior @ H.T @ np.linalg.inv(H @ P_prior @ H.T + (Omega + R) / alpha ) @ H @ P_prior

    return x_hat, P_hat


# def rplf_update_backup(x_prior, P_prior, R, z):
#     # beta = 0.01
#     gamma = 0.01
#
#     x_hat = x_prior
#     P_hat = P_prior
#     x_hat_old = 100 * np.ones_like(x_prior)
#     P_hat_old = 100 * np.eye(4)
#     ite = 0
#     while kl_divergence_gaussian(x_hat_old, P_hat_old, x_hat, P_hat) >= 1e-2 and ite <= 100:
#         ite += 1
#         x_hat_old = x_hat
#         P_hat_old = P_hat
#         points_linear = MerweScaledSigmaPoints(n=4, alpha=1e-3, beta=2., kappa=0.)
#         sigmas = points_linear.sigma_points(x_hat, P_hat)
#         sigmas_h = np.empty((9, 2))
#         for i in range(sigmas.shape[0]):
#             sigmas_h[i] = h(sigmas[i])
#         zp, Pz = unscented_transform(sigmas_h, points_linear.Wm, points_linear.Wc, noise_cov=None)
#         Pxz = np.zeros((4, 2))
#         for i in range(sigmas.shape[0]):
#             Pxz += points_linear.Wc[i] * np.outer(sigmas[i] - x_hat, sigmas_h[i] - zp)
#
#         H = Pxz.T @ np.linalg.inv(P_hat)
#         b = zp - H @ x_hat
#         Omega = Pz - H @ P_hat @ H.T
#
#         # x_hat = x_prior + (beta + 1) / (
#         #         (2 * np.pi) ** (2 * beta / 2) * (np.linalg.det(R)) ** (beta / 2)) * P_prior @ H.T @ np.linalg.inv(
#         #     H @ P_prior @ H.T + Omega + R) @ (z - H @ x_prior - b)
#         # P_hat = P_prior - P_prior @ H.T @ np.linalg.inv(H @ P_prior @ H.T + (Omega + R) * (
#         #         (2 * np.pi) ** (2 * beta / 2) * (np.linalg.det(R)) ** (beta / 2)) / (beta + 1)) @ H @ P_prior
#
#         x_hat = x_prior + (gamma + 1) ** ((3 * gamma + 2) / (2 + 2 * gamma)) * (2 * np.pi) ** (-gamma / (1 + gamma)) * (
#             np.linalg.det(R)) ** ((-gamma) / (2 + 2 * gamma)) * P_prior @ H.T @ np.linalg.inv(
#             H @ P_prior @ H.T + Omega + R) @ (z - H @ x_prior - b)
#         P_hat = P_prior - P_prior @ H.T @ np.linalg.inv(H @ P_prior @ H.T + (Omega + R) / (
#                 (gamma + 1) ** ((3 * gamma + 2) / (2 + 2 * gamma)) * (2 * np.pi) ** (-gamma / (1 + gamma)) * (
#             np.linalg.det(R)) ** ((-gamma) / (2 + 2 * gamma)))) @ H @ P_prior
#
#         # x_hat = x_prior + 0.8* P_prior @ H.T @ np.linalg.inv(
#         #     H @ P_prior @ H.T + Omega + R) @ (z - H @ x_prior - b)
#         # P_hat = P_prior - P_prior @ H.T @ np.linalg.inv(H @ P_prior @ H.T + (Omega + R) / (0.8) ) @ H @ P_prior
#     return x_hat, P_hat


# def rplf_update(x_prior, P_prior, R, z):
#     beta = 0.01
#
#     def l(x):
#         loss = -(beta + 1) / (beta * (2 * np.pi) ** (2 * beta / 2) * (np.linalg.det(R)) ** (beta / 2)) * np.exp(
#             -beta / 2 * (z - h(x)).T @ np.linalg.inv(R) @ (z - h(x)))
#         return loss
#
#     x_hat = x_prior
#     P_hat = P_prior
#     x_hat_old = 100 * np.ones_like(x_prior)
#     P_hat_old = 100 * np.eye(4)
#     ite = 0
#     while kl_divergence_gaussian(x_hat_old, P_hat_old, x_hat, P_hat) >= 1e-2 and ite <= 5:
#         ite += 1
#         x_hat_old = x_hat
#         P_hat_old = P_hat
#
#         points_linear = JulierSigmaPoints(n=4, kappa=0.)
#         sigmas = points_linear.sigma_points(x_hat, P_hat)
#         sigmas_J = np.empty((9, 4))
#         sigmas_H = np.empty((9, 4, 4))
#
#         jacobian_l = jacobian(l)
#         hessian_l = hessian(l)
#         for i in range(sigmas.shape[0]):
#             sigmas_J[i] = jacobian_l(sigmas[i])
#             sigmas_H[i] = hessian_l(sigmas[i])
#
#         x_hat_update = np.dot(points_linear.Wm, sigmas_J)
#         # P_hat_update = np.zeros((4, 4))
#         # for i in range(sigmas.shape[0]):
#         #     P_hat_update += points_linear.Wm[i] * sigmas_H[i]
#
#         sigmas_h = np.empty((9, 2))
#         for i in range(sigmas.shape[0]):
#             sigmas_h[i] = h(sigmas[i])
#         zp, Pz = unscented_transform(sigmas_h, points_linear.Wm, points_linear.Wc, noise_cov=None)
#         Pxz = np.zeros((4, 2))
#         for i in range(sigmas.shape[0]):
#             Pxz += points_linear.Wc[i] * np.outer(sigmas[i] - x_hat, sigmas_h[i] - zp)
#
#         H = Pxz.T @ np.linalg.inv(P_hat)
#         b = zp - H @ x_hat
#         Omega = Pz - H @ P_hat @ H.T
#
#         x_hat = x_prior - P_prior @ x_hat_update
#         P_hat = P_prior - P_prior @ H.T @ np.linalg.inv(H @ P_prior @ H.T + (Omega + R) * (
#                 (2 * np.pi) ** (2 * beta / 2) * (np.linalg.det(R)) ** (beta / 2)) / (beta + 1)) @ H @ P_prior
#
#     return x_hat, P_hat


# 创建sigma点
points1 = MerweScaledSigmaPoints(n=4, alpha=1e-3, beta=2., kappa=0.)
points2 = MerweScaledSigmaPoints(n=4, alpha=1e-3, beta=2., kappa=0.)
points3 = MerweScaledSigmaPoints(n=4, alpha=1e-3, beta=2., kappa=0.)

# 创建UKF滤波器
ukf = UKF(dim_x=4, dim_z=2, dt=0.1, fx=f, hx=h, points=points1)
ukf.x = np.array([0., 0., 1., 1.])
ukf.P *= 1
ukf.R = np.diag([0.01, 0.01])
ukf.Q = np.diag([1, 1, 0.01, 0.01])

plf = UKF(dim_x=4, dim_z=2, dt=0.1, fx=f, hx=h, points=points2)
plf.x = np.array([0., 0., 1., 1.])
plf.P *= 1
plf.R = np.diag([0.01, 0.01])
plf.Q = np.diag([1, 1, 0.01, 0.01])

rplf = UKF(dim_x=4, dim_z=2, dt=0.1, fx=f, hx=h, points=points3)
rplf.x = np.array([0., 0., 1., 1.])
rplf.P *= 1
rplf.R = np.diag([0.01, 0.01])
rplf.Q = np.diag([1, 1, 0.01, 0.01])

# 生成真实状态和观测数据
num_steps = 100
true_states = []
observations = []
state = np.array([0., 0., 1., 1.])
for _ in range(num_steps):
    state = f(state, 0.1) + np.random.multivariate_normal(mean=[0, 0, 0, 0], cov=ukf.Q)  # 加入一些噪声
    true_states.append(state)

    observation = h(state) + 0.9 * np.random.multivariate_normal(mean=[0, 0],
                                                                 cov=ukf.R) + 0.1 * np.random.multivariate_normal(
        mean=[0, 0], cov=10 * ukf.R)  # 生成观测值并加入噪声

    observations.append(observation)

true_states = np.array(true_states)

# 使用UKF进行状态估计
xs_ukf = []  # 保存估计的状态
xs_plf = []
xs_rplf = []
for z in observations:
    ukf.predict()
    ukf.update(z)

    plf.predict()
    x_pos_plf, P_pos_plf = plf_update(plf.x, plf.P, plf.R, z)
    plf.x = x_pos_plf.copy()
    plf.P = P_pos_plf.copy()

    rplf.predict()
    x_pos_rplf, P_pos_rplf = rplf_update(rplf.x, rplf.P, rplf.R, z)
    rplf.x = x_pos_rplf.copy()
    rplf.P = P_pos_rplf.copy()

    xs_ukf.append(ukf.x.copy())
    xs_plf.append(plf.x.copy())
    xs_rplf.append(rplf.x.copy())

xs_ukf = np.array(xs_ukf)
xs_plf = np.array(xs_plf)
xs_rplf = np.array(xs_rplf)
observations = np.array(observations)

labels = ['p_x', 'p_y', 'v_x', 'v_y']
plt.figure(figsize=(12, 8))
for i in range(4):
    plt.subplot(2, 2, i + 1)
    plt.plot(xs_ukf[:, i], label='x_ukf')
    plt.plot(xs_plf[:, i], label='x_plf')
    plt.plot(xs_rplf[:, i], label='x_rplf')
    plt.plot(true_states[:, i], label='x_true', linestyle='--')
    if i < 2:
        plt.plot(observations[:, i], label='y')
    plt.title(labels[i])
    plt.legend()

plt.tight_layout()
plt.show()


def compute_rmse(predictions, true_values):
    """
    Compute the Root Mean Square Error (RMSE) between predictions and true values.
    """
    return np.sqrt(((predictions - true_values) ** 2).mean(axis=0))


# Compute RMSE for each method
rmse_ukf = compute_rmse(xs_ukf, true_states)
rmse_plf = compute_rmse(xs_plf, true_states)
rmse_rplf = compute_rmse(xs_rplf, true_states)
print("rmse_ukf", rmse_ukf)
print("rmse_plf", rmse_plf)
print("rmse_rplf", rmse_rplf)



import autograd.numpy as np
from autograd import jacobian, hessian
from autograd.numpy import sin, cos, arctan, pi, arctan2

landmarks = np.array([
    [0, 0], [0, 25], [0, 50], [0, 75], [0, 100],
    [25, 0], [25, 25], [25, 50], [25, 75], [25, 100],
    [50, 0], [50, 25], [50, 50], [50, 75], [50, 100],
    [75, 0], [75, 25], [75, 50], [75, 75], [75, 100],
    [100, 0], [100, 25], [100, 50], [100, 75], [100, 100],
])

class Sensor_Network:

    dt : float = 0.5
    def __init__(self, state_outlier_flag=False, 
                measurement_outlier_flag=False, noise_type='Gaussian'):
        dt = self.dt
        self.F = np.array([[1, 0, dt, 0],
                           [0, 1, 0, dt],
                           [0, 0, 1, 0],
                           [0, 0, 0, 1],
                           ])

        self.dim_x = 4
        self.dim_y = 25
        self.P0 = np.diag(np.array([49, 1, 4, 2]))
        self.x0 = np.random.multivariate_normal(mean=np.zeros(self.dim_x), cov=self.P0)

        self.state_outlier_flag = state_outlier_flag
        self.measurement_outlier_flag = measurement_outlier_flag
        self.noise_type = noise_type
        self.obs_var = np.full(25, 0.1)
        self.Q = 0.04 * np.array([
            [dt**3/3, 0, dt**2/2, 0],
            [0, dt**3/3, 0, dt**2/2],
            [dt**2/2, 0, dt, 0],
            [0, dt**2/2, 0, dt]
        ])
        self.R = np.diag(self.obs_var)


    def f(self, x, dt=None, u=None):
        if u is None:
            return self.F @ x

    def h(self, x):
        W0 = 1000
        d0 = 1
        hx = []
        px, py, vx, vy = x
        for i in range(25):
            dist = (px - landmarks[i][0])**2 + (py - landmarks[i][1])**2
            if dist <= d0:
                hx.append(np.sqrt(W0))
            else:
                hx.append(np.sqrt(W0*d0/dist))

        return np.array(hx)

    def jac_f(self, x_hat):
        return self.F

    def jac_h(self, x_hat):
        return jacobian(lambda x: self.h(x))(x_hat)

    def f_withnoise(self, x):
        if self.state_outlier_flag:
            prob = np.random.rand()
            if prob <= 0.9:
                cov = self.Q  # 95%概率使用Q
            else:
                cov = 100 * self.Q  # 5%概率使用100Q
        else:
            cov = self.Q
        return self.f(x) + np.random.multivariate_normal(mean=np.zeros(self.dim_x), cov=cov)
    
    def h_withnoise(self, x):
        if self.noise_type == 'Gaussian':
            if self.measurement_outlier_flag:
                prob = np.random.rand()
                if prob <= 0.9:
                    cov = self.R  # 95%概率使用R
                else:
                    cov = 1000 * self.R  # 5%概率使用100R
            else:
                cov = self.R
            return self.h(x) + np.random.multivariate_normal(mean=np.zeros(self.dim_y), cov=cov)
        else:
            return self.h(x) + np.random.laplace(loc=0, scale=1, size=(self.dim_y, ))


import autograd.numpy as np
from autograd import jacobian, hessian
from autograd.numpy import sin, cos, arctan, pi, arctan2, sqrt
from data_processing import load_data
from .model import Model

class UGV(Model):

    dt : float = 1.0 / 15

    def __init__(self, state_outlier_flag=False, 
                measurement_outlier_flag=False, noise_type='Gaussian'):
        super().__init__(self)
        self.dim_x = 3
        self.dim_y = 6
        self.x0 = np.array([0., 0., 0.])
        self.P0 = np.diag(np.array([0.0001, 0.0001, 0.0001])) ** 2

        self.state_outlier_flag = state_outlier_flag
        self.measurement_outlier_flag = measurement_outlier_flag
        self.noise_type = noise_type
        self.alpha = 2.0
        self.beta = 5.0

        self.process_std = np.array([0.0034, 0.0056, 0.0041])
        self.observation_std = np.array([0.0238, 0.0284, 0.0259, 0.0107, 0.0094, 0.0118])
        self.obs_var = np.ones(self.dim_y) * 0.01
        self.Q = np.diag(self.process_std**2)
        self.R = np.diag(self.observation_std**2)

    def f(self, x, u, dt=None):
        if dt is None:
            dt = self.dt
        x0, x1, x2 = x
        x0_ = x0 + dt * u[0] * np.cos(x2) - 0.0001675046729610055
        x1_ = x1 + dt * u[0] * np.sin(x2) - 0.0001963914687308423
        x2_ = x2 + dt * u[1] + 0.0005640178926637775
        return np.array([x0_, x1_, x2_])

    def h(self, x):
        px, py, theta = x
        obstacle_info = [[1.052, -2.695], [4.072, -1.752], [6.028, -3.324]]
        obstacle = np.array(obstacle_info)
        der_x_robot = 0.329578
        rot = np.array([[np.cos(theta), np.sin(theta)], [-np.sin(theta), np.cos(theta)]])
        dist = []
        obstacle_obs = []
        for M in obstacle:
            obstacle_obs.append(((M - x[:2])@rot.T) - np.array([der_x_robot, 0]))
        obstacle_obs = np.array(obstacle_obs)
        dist = np.linalg.norm(obstacle_obs, axis=1)
        angle = np.arctan2(obstacle_obs[:, 1], obstacle_obs[:, 0])
        return np.concatenate([dist, angle]) + np.array([-0.0312, -0.0581, -0.0557, 0.0053, 0.0059, 0.0125])
        

    def jac_f(self, x, u):
        px, py, theta = x
        v, w = u
        dt = self.dt
        return np.array([
            [1, 0, -v*sin(theta)*dt],
            [0, 1, v*cos(theta)*dt],
            [0, 0, 1]
        ])

    def jac_h(self, x_hat):
        return jacobian(lambda x: self.h(x))(x_hat)
    
    def get_sensor_data(self, filepath_list, min_len):
        state = []
        action = []
        obs = []
        for file_path in filepath_list:
            state_, action_, obs_ = load_data(file_path, min_len)
            state += state_
            action += action_
            obs += obs_
        NUM_RUNS = len(state)
        X_list = []
        Y_list = []
        U_list = []
        for i in range(NUM_RUNS):
            X_list.append(np.array(state[i]))
            Y_list.append(np.array(obs[i]))
            U_list.append(np.array(action[i]))
        X_arr = np.array(X_list)
        Y_arr = np.array(Y_list)
        U_arr = np.array(U_list)
        zero_states = X_arr[:, 0, :]
        return X_arr, Y_arr, U_arr, zero_states
    
    def f_withnoise(self, x, u=None):
        if self.state_outlier_flag:
            prob = np.random.rand()
            if prob <= 0.95:
                cov = self.Q  # 95%概率使用Q
            else:
                cov = 100 * self.Q  # 5%概率使用100Q
        else:
            cov = self.Q
        return self.f(x, u) + np.random.multivariate_normal(mean=np.zeros(self.dim_x), cov=cov)
    
    def h_withnoise(self, x):
        if self.noise_type == 'Gaussian':
            if self.measurement_outlier_flag:
                prob = np.random.rand()
                if prob <= 0.99:
                    cov = self.R  # 95%概率使用R
                else:
                    cov = 10 * self.R  # 5%概率使用100R
            else:
                cov = self.R
            return self.h(x) + np.random.multivariate_normal(mean=np.zeros(self.dim_y), cov=cov)
        elif self.noise_type == 'Beta':
            noise = np.random.beta(self.alpha, self.beta, self.dim_y)
            noise = noise - np.mean(noise)
            return self.h(x) + noise
        else:
            return self.h(x) + np.random.laplace(loc=0, scale=self.obs_var, size=(self.dim_y, ))



from dataclasses import dataclass
import autograd.numpy as np
from autograd import jacobian, hessian
from autograd.numpy import sin, cos, arctan, pi, arctan2

class ChemicalReaction:

    dt : float = 0.10

    def __init__(self, state_outlier_flag=False, measurement_outlier_flag=False):

        self.dim_x = 2
        self.dim_y = 1
        self.x0 = np.array([0.1, 4.5])
        self.P0 = 1e-4 * np.eye(self.dim_x)

        self.state_outlier_flag = state_outlier_flag
        self.measurement_outlier_flag = measurement_outlier_flag
        self.Q = 1e-4 * np.eye(self.dim_x)
        self.R = 1e-2 * np.eye(self.dim_y)

    def f(self, x, dt=None):
        # reaction parmater
        k1 = 0.16
        k2 = 0.0064
        if dt is None:
            dt = self.dt
        x1, x2 = x
        x1_ = x1 + (-2 * k1 * x1 * x1 + 2 * k2 * x2) * dt
        x2_ = x2 + (k1 * x1 * x1 - k2 * x2) * dt
        # print('state:', np.array([x1_, x2_]))
        return np.array([x1_, x2_])

    def h(self, x):
        x1, x2 = x
        # print('obs:', np.array([x1 + x2]))
        return np.array([x1 + x2])
    
    def f_withnoise(self, x):
        if self.state_outlier_flag:
            prob = np.random.rand()
            if prob <= 0.95:
                cov = self.Q  # 95%概率使用Q
            else:
                cov = 100 * self.Q  # 5%概率使用100Q
        else:
            cov = self.Q
        return self.f(x) + np.random.multivariate_normal(mean=np.zeros(self.dim_x), cov=cov)
    
    def h_withnoise(self, x):
        if self.measurement_outlier_flag:
            prob = np.random.rand()
            if prob <= 0.8:
                cov = self.R  # 95%概率使用R
            else:
                cov = 1000 * self.R  # 5%概率使用100R
        else:
            cov = self.R
        return self.h(x) + np.random.multivariate_normal(mean=np.zeros(self.dim_y), cov=cov)

    def jac_f(self, x):
        k1 = 0.16
        k2 = 0.0064
        x1, x2 = x
        jac_11 = 1 - 4 * k1 * x1 * self.dt
        jac_12 = -2 * k2 * self.dt
        jac_21 = 2 * k1 * x1 * self.dt
        jac_22 = 1 - k2 * self.dt

        # print('state:', np.array([x1_, x2_]))
        return np.array([[jac_11, jac_12], [jac_21, jac_22]])

    def jac_h(self, x):
        jac_h11 = 1
        jac_h12 = 1
        return np.array([[jac_h11, jac_h12]])


import autograd.numpy as np
import warnings

def is_positive_semidefinite(matrix):
    if matrix.shape[0] != matrix.shape[1]:
        raise ValueError('Matrix is not square')
    
    eigenvalues = np.linalg.eigvals(matrix)
    # print('Eigenvalues: ', eigenvalues)
    
    if np.any(eigenvalues < 0):
        print('Eigenvalues: ', eigenvalues)
        warnings.warn('Matrix is not positive semidefinite')
        
            
def cal_mean(func, mean, var, points):
    sigmas = points.sigma_points(mean, var)
    first_eval = func(sigmas[0])
    
    if isinstance(first_eval, np.ndarray):
        sigmas_func = np.zeros((sigmas.shape[0], *first_eval.shape))
    else:
        sigmas_func = np.zeros(sigmas.shape[0])
    
    for i, s in enumerate(sigmas):
        sigmas_func[i] = func(s)

        # if isinstance(sigmas_func[i], np.ndarray) and sigmas_func[i].ndim == 2:
        #     is_positive_semidefinite(sigmas_func[i])
    
    mean_func = np.tensordot(points.Wm, sigmas_func, axes=([0], [0]))
    return mean_func

def cal_mean_mc(func, mean, var, num_samples=10):
    samples = np.random.multivariate_normal(mean, var, num_samples)
    # 计算每个样本在函数 f 上的值
    sample_values = np.apply_along_axis(func, 1, samples)
    # print('1:', sample_values.shape)
    # 计算样本值的平均值
    expectation = np.mean(sample_values, axis=0)
    # print('2:', expectation.shape)
    return expectation

def kl_divergence(mean0, cov0, mean1, cov1):
    # 计算维数
    k = mean0.shape[0]
    
    # 计算协方差矩阵的逆
    cov1_inv = np.linalg.inv(cov1)
    
    # 计算两个分布均值之差
    mean_diff = mean1 - mean0
    
    # 计算公式的每一部分
    term1 = np.trace(np.dot(cov1_inv, cov0))
    term2 = np.dot(np.dot(mean_diff.T, cov1_inv), mean_diff)
    term3 = np.log(np.linalg.det(cov1) / np.linalg.det(cov0))
    
    # 计算 KL 散度
    kl_div = 0.5 * (term1 + term2 - k + term3)
    
    return kl_div


from einops import rearrange, reduce
from typing import Dict
from filterpy.kalman import UnscentedKalmanFilter, MerweScaledSigmaPoints
from filterpy.kalman import unscented_transform as UT
import autograd.numpy as np
from autograd.numpy import eye, ones, zeros, dot, isscalar, outer
from copy import deepcopy

class UKF(UnscentedKalmanFilter):
    
    def __init__(self, model):
        super().__init__(
            fx = model.f,
            hx = model.h,
            dt = model.dt,
            dim_x = model.dim_x,
            dim_z = model.dim_y,
            points = MerweScaledSigmaPoints(model.dim_x, alpha=0.1, beta=2.0, kappa=1.0)
        )

        self.Q = model.Q
        self.R = model.R

        self.x = model.x0
        self.P = model.P0
        
    def compute_process_sigmas(self, u=0, dt=None, fx=None):

        if fx is None:
            fx = self.fx
        # calculate sigma points for given mean and covariance
        sigmas = self.points_fn.sigma_points(self.x, self.P)

        for i, s in enumerate(sigmas):
            self.sigmas_f[i] = fx(s, u)
        
    def predict(self, u=0):
        self.compute_process_sigmas(u)
        #and pass sigmas through the unscented transform to compute prior
        self.x, self.P = UT(self.sigmas_f, self.Wm, self.Wc, self.Q,
                        self.x_mean, self.residual_x)
        self.sigmas_f = self.points_fn.sigma_points(self.x, self.P)
        
        self.x_prior = np.copy(self.x)
        self.P_prior = np.copy(self.P)
    
    def update(self, z):
        # z = self.get_y(wk, phik, dphik, yawk)
        sigmas_h = []
        for s in self.sigmas_f:
            sigmas_h.append(self.hx(s))
        self.sigmas_h = np.atleast_2d(sigmas_h)
        zp, self.S = UT(self.sigmas_h, self.Wm, self.Wc, self.R, self.z_mean, self.residual_z)
        self.SI = self.inv(self.S)
        Pxz = self.cross_variance(self.x, zp, self.sigmas_f, self.sigmas_h)
        self.K = dot(Pxz, self.SI)        # Kalman gain
        self.y = self.residual_z(z, zp)   # residual

        self.x = self.x + dot(self.K, self.y)
        self.P = self.P - dot(self.K, dot(self.S, self.K.T))

        # save measurement and posterior state
        self.z = deepcopy(z)
        self.x_post = self.x.copy()
        self.P_post = self.P.copy()


from .air_traffic import Air_Traffic
from .blackhole import BlackHole
from .chemical_reaction import ChemicalReaction
from .lorenz import Lorenz
from .robot_move import RobotMove
from .sensor_net import Sensor_Network
from .sin_cos import SinCos
from .slide_slip_veh import Vehicle
from .ugv import UGV
from .ugv2 import UGV2
from .wiener_velocity import WienerVelocity



from filterpy.kalman import ExtendedKalmanFilter
import autograd.numpy as np


class EKF(ExtendedKalmanFilter):
    def __init__(self, model):
        super().__init__(dim_x=model.dim_x,
                         dim_z=model.dim_y,
                         )
        
        self.f = model.f
        self.h = model.h
        self.jac_f = model.jac_f
        self.jac_h = model.jac_h

        self.Q = model.Q
        self.R = model.R

        self.x = model.x0
        self.P = model.P0 

    def predict(self, u=0):
        F = self.jac_f(self.x, u)
        self.x = self.f(self.x, u)
        self.P = F @ self.P @ F.T + self.Q
        
        self.x_prior = self.x.copy()
        self.P_prior = self.P.copy()

    def update(self, y):
        H = self.jac_h(self.x)
        hx = self.h(self.x)
        PHT = self.P @ H.T
        self.S = H @ PHT + self.R
        self.K = PHT @ np.linalg.inv(self.S)
        self.x = self.x + self.K @ (y - hx)
        I_KH = self._I - self.K @ H
        self.P = (I_KH @ self.P @ I_KH.T) + (self.K @ self.R @ self.K.T)
        self.x_post = self.x.copy()
        self.P_post = self.P.copy()


import numpy as np
from filterpy.monte_carlo import systematic_resample
from numpy.random import uniform
from scipy.stats import multivariate_normal,norm


class PF:
    def __init__(self, model, num_particles=100):

        self.f = model.f
        self.h = model.h
        self.dim_x=model.dim_x
        self.dim_y=model.dim_y

        self.Q = model.Q
        self.R = model.R
        self.x = model.x0
        self.P = model.P0
        self.ex = np.zeros(self.dim_x)
        self.ey = np.zeros(self.dim_y)

        self.num_particles = num_particles
        self.particles = np.random.multivariate_normal(mean=self.x, cov=self.P,
                                                       size=self.num_particles)
        self.weights = np.ones(num_particles) / num_particles
        self.x_prior = self.x.copy()
        self.x_post = self.x.copy()

    def predict(self):
        for i in range(self.num_particles):
            self.particles[i] = self.f(self.particles[i]) + np.random.multivariate_normal(self.ex,
                                                                                          self.Q)
        self.x_prior = np.average(self.particles, weights=self.weights, axis=0)
        self.x = self.x_prior.copy()

    def update(self, y):

        for i in range(self.num_particles):
            y_pred = self.h(self.particles[i])
            self.weights[i] *= multivariate_normal.pdf(y, y_pred, self.R)

        self.weights += 1.e-300
        self.weights /= sum(self.weights)

        # 2. 重采样
        if self.neff(self.weights) < self.num_particles / 2:
            indexes = systematic_resample(self.weights)
            self.resample(indexes)
        self.x_post = np.average(self.particles, weights=self.weights, axis=0)
        self.x = self.x_post.copy()

    def neff(self, weights):
        return 1. / np.sum(np.square(weights))

    def resample(self, indexes):
        self.particles[:] = self.particles[indexes]
        self.weights.resize(len(self.particles))
        self.weights.fill(1.0 / len(self.weights))




import autograd.numpy as np

class Model:

    def __init__(self, state_outlier_flag=False, 
                measurement_outlier_flag=False, noise_type='Gaussian'):

        self.state_outlier_flag = state_outlier_flag
        self.measurement_outlier_flag = measurement_outlier_flag
        self.noise_type = noise_type
    
    def f(self, x, u=None):
        pass

    def h(self, x):
        pass

    def jac_f(self, x, u=None):
        pass

    def jac_h(self, x):
        pass

    def f_withnoise(self, x, u=None):
        pass

    def h_withnoise(self, x):
        pass
    
    


from einops import rearrange, reduce
from typing import Dict
from filterpy.kalman import UnscentedKalmanFilter, MerweScaledSigmaPoints
from filterpy.kalman import unscented_transform as UT
import autograd.numpy as np
from autograd.numpy import eye, ones, zeros, dot, isscalar, outer
from copy import deepcopy
from .utils import is_positive_semidefinite, kl_divergence

class PLF(UnscentedKalmanFilter):
    
    def __init__(self, model):
        super().__init__(
            fx = model.f,
            hx = model.h,
            dt = model.dt,
            dim_x = model.dim_x,
            dim_z = model.dim_y,
            points = MerweScaledSigmaPoints(model.dim_x, alpha=0.1, beta=2.0, kappa=1.0)
        )

        self.Q = model.Q
        self.R = model.R

        self.x = model.x0
        self.P = model.P0
        
    def compute_process_sigmas(self, u=0, dt=None, fx=None):

        if fx is None:
            fx = self.fx
        # calculate sigma points for given mean and covariance
        sigmas = self.points_fn.sigma_points(self.x, self.P)

        for i, s in enumerate(sigmas):
            self.sigmas_f[i] = fx(s, u)
        
    def predict(self, u=0):
        self.compute_process_sigmas(u)
        #and pass sigmas through the unscented transform to compute prior
        self.x, self.P = UT(self.sigmas_f, self.Wm, self.Wc, self.Q,
                        self.x_mean, self.residual_x)
        
        self.x_prior = np.copy(self.x)
        self.P_prior = np.copy(self.P)
    
    def update(self, z):
        x_hat = self.x_prior
        P_hat = self.P_prior
        x_hat_old = 100 * np.ones_like(x_hat)
        P_hat_old = 100 * np.eye(self._dim_x)
        iter = 0
        while kl_divergence(x_hat_old, P_hat_old, x_hat, P_hat) >= 1e-4 and iter <= 100:
            iter += 1
            x_hat_old = x_hat
            P_hat_old = P_hat
            sigmas = self.points_fn.sigma_points(x_hat, P_hat)
            sigmas_h = zeros((self._num_sigmas, self._dim_z))
            for i, s in enumerate(sigmas):
                sigmas_h[i] = self.hx(s)
            zp, Pz = UT(sigmas_h, self.Wm, self.Wc, noise_cov=None)
            Pxz = zeros((self._dim_x, self._dim_z))
            for i in range(sigmas.shape[0]):
                Pxz += self.Wc[i] * np.outer(sigmas[i] - x_hat, sigmas_h[i] - zp)
            H = Pxz.T @ np.linalg.inv(P_hat)
            b = zp - H @ x_hat
            Omega = Pz - H @ P_hat @ H.T

            x_hat = self.x_prior + self.P_prior @ H.T @ np.linalg.inv(H @ self.P_prior @ H.T + Omega + self.R) @ (z - H @ self.x_prior - b)
            P_hat = self.P_prior - self.P_prior @ H.T @ np.linalg.inv(H @ self.P_prior @ H.T + Omega + self.R) @ H @ self.P_prior
        
        self.x = x_hat
        self.P = P_hat
        self.x_post = self.x.copy()
        self.P_post = self.P.copy()


from setuptools import setup, find_packages

# 读取 README 文件作为 long_description
with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="NANO-filter",  # 包名
    version="1.0.0",  # 版本号
    license="MIT",
    author="Tianyi Zhang",
    author_email="zhangtia24@mails.tsinghua.edu.cn",
    description="Nonlinear Bayesian Filtering with Natural Gradient Gaussian Approximation",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/TianyiMehdi/NANO-filter",  # GitHub仓库地址
    packages=find_packages(),  # 自动查找包
    install_requires=[
        "autograd==1.6.2",
        "dataclasses_json==0.6.7",
        "einops==0.8.0",
        "filterpy==1.4.5",
        "matplotlib==3.8.2",
        "numpy==2.1.2",
        "pandas==2.2.3",
        "scipy==1.14.1",
        "seaborn==0.13.2",
        "tqdm==4.66.1",
    ],  # 安装依赖
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",  # 更改为你的许可证
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.8",  # Python版本要求
)



from dataclasses import dataclass
import autograd.numpy as np
from autograd import jacobian, hessian
from autograd.numpy import sin, cos, arctan


@dataclass
class Vehicle:
    """Vehicle config for the slide-slip vehicle model.
    """
    # Vehicle parameters
    m: float = 1500 # mass [kg]
    dt: float = 0.01
    B: float = 14.0
    C: float = 1.43
    D: float = 0.75
    u: float = 20.0
    I_zz: float = 2420.0
    L: float = 2.54
    a: float = 1.14
    b: float = 1.40
    g: float = 9.81

    def __init__(self, state_outlier_flag=False, 
                measurement_outlier_flag=False, noise_type='Gaussian'):
        D = self.D
        g = self.g
        dt = self.dt
        u = self.u
        L = self.L
        m = self.m
        a = self.a
        b = self.b
        I_zz = self.I_zz
        self.A1 = -D*g*dt / (u*L)
        self.A2 = D*m*g*a*b*dt / (I_zz * L)

        self.dim_x = 2
        self.dim_y = 2
        self.P0 = 0.1 * np.eye(self.dim_x)
        self.x0 =  np.random.multivariate_normal(mean=np.zeros(self.dim_x), cov=self.P0)


        self.state_outlier_flag = state_outlier_flag
        self.measurement_outlier_flag = measurement_outlier_flag
        self.noise_type = noise_type
        self.var = np.array([1e-2, 1e-2])
        self.obs_var = np.array([1e-2, 1e-2])
        self.Q = np.diag(self.var)
        self.R = np.diag(self.obs_var)
    
    def f(self, x):
        """Transition function for the slide-slip vehicle model.
        states: (2, )
        returns: (2, )
        """
        A1 = self.A1
        b = self.b
        C = self.C
        B = self.B
        a = self.a
        u = self.u
        b = self.b
        A2 = self.A2
        dt = self.dt

        theta = x[0]
        ut = - theta  # control policy
        # ut = jnp.zeros_like(theta)
        omega = x[1]
        v1 = sin(C*arctan(B*(theta - ut + a*omega/u)))
        v2 = sin(C*arctan(B*(theta - b*omega/u)))
        f1 = A1 * b * cos(ut) * v1 + A1 * a * v2 - omega * dt
        f2 = -A2 * v1 + A2 * v2
        return np.stack([f1, f2])
    
    def h(self, x):
        theta, omega = x
        alpha_1 = theta + theta + omega * self.a / self.u
        alpha_2 = theta - omega * self.b / self.u
        FY1 = - sin(self.C * arctan(self.B * alpha_1))
        FY2 = - sin(self.C * arctan(self.B * alpha_2))
        return np.array([FY1, FY2])
    
    def f_withnoise(self, x):
        if self.state_outlier_flag:
            prob = np.random.rand()
            if prob <= 0.95:
                cov = self.Q  # 95%概率使用Q
            else:
                cov = 100 * self.Q  # 5%概率使用100Q
        else:
            cov = self.Q
        return self.f(x) + np.random.multivariate_normal(mean=np.zeros(self.dim_x), cov=cov)
    
    def h_withnoise(self, x):
        if self.noise_type == 'Gaussian':
            if self.measurement_outlier_flag:
                prob = np.random.rand()
                if prob <= 0.95:
                    cov = self.R  # 95%概率使用R
                else:
                    cov = 100 * self.R  # 5%概率使用100R
            else:
                cov = self.R
            return self.h(x) + np.random.multivariate_normal(mean=np.zeros(self.dim_y), cov=cov)
        else:
            return self.h(x) + np.random.laplace(loc=0, scale=1, size=(self.dim_y, ))

    def jac_f(self, x_hat):
        return jacobian(lambda x: self.f(x))(x_hat)
    
    def jac_h(self, x_hat):
        return jacobian(lambda x: self.h(x))(x_hat)

if __name__ == '__main__':
    x0 = np.random.uniform(-0.05, 0.05, size=(2, ))
    vehicle = Vehicle()
    print(vehicle.f(x0), vehicle.h(x0))
    print(vehicle.jac_f(x0))
    print(vehicle.jac_h(x0))

# # generate dataset
# def transition_numpy(states: np.ndarray, vehicle: Vehicle):
#     """Transition function for the slide-slip vehicle model.
#     states: (num_traj, 2)
#     returns: (num_traj, 2)
#     """
#     A1 = vehicle.A1
#     b = vehicle.b
#     C = vehicle.C
#     B = vehicle.B
#     a = vehicle.a
#     u = vehicle.u
#     b = vehicle.b
#     A2 = vehicle.A2
#     dt = vehicle.dt

#     theta = states[:, 0]
#     ut = - theta  # control policy
#     omega = states[:, 1]
#     v1 = np.sin(C*np.arctan(B*(theta - ut + a*omega/u)))
#     v2 = np.sin(C*np.arctan(B*(theta - b*omega/u)))
#     f1 = A1 * b * np.cos(ut) * v1 + A1 * a * v2 - omega * dt
#     f2 = -A2 * v1 + A2 * v2
#     return np.stack([f1, f2], axis=1)

# def process_noises(size: int, var: np.ndarray) -> np.ndarray:
#     return np.random.multivariate_normal(mean=np.zeros(2), cov=np.diag(var), size=size).astype(np.float32)

# def observation_noises(size: int, obsvar: np.ndarray) -> np.ndarray:
#     return np.random.multivariate_normal(mean=np.zeros(2), cov=np.diag(obsvar), size=size).astype(np.float32)

# def generate_data(num_traj: int, traj_len: int, save_dir: str, mode:str, vehicle: Vehicle, seed: int = 42):
#     np.random.seed(seed)
#     data_dir = Path(save_dir)
#     data_dir.mkdir(parents=True, exist_ok=True)
#     save_path = data_dir / f"{mode}.npz"

#     states = np.zeros((num_traj, traj_len, 2), dtype=np.float32)
#     observations = np.zeros((num_traj, traj_len, 2), dtype=np.float32)
#     states[:, 0, :] = np.random.uniform(-0.05, 0.05, size=(num_traj, 2))
#     for i in tqdm(range(traj_len - 1)):
#         states[:, i+1, :] = vehicle.f(states[:, i, :]) + process_noises(num_traj, vehicle.var)
#     observations = states + observation_noises((num_traj, traj_len), vehicle.obs_logvar)
#     np.savez_compressed(save_path, states=states, observations=observations)


# if __name__ == "__main__":
#     parser = ArgumentParser()
#     parser.add_argument("--num_traj", type=int, default=30)
#     parser.add_argument("--traj_len", type=int, default=50)
#     parser.add_argument("--save_dir", type=str, default="dataset")
#     parser.add_argument("--mode", type=str, default="vehicle")
#     parser.add_argument("--seed", type=int, default=42)
#     args = parser.parse_args()
#     vehicle = Vehicle()
#     generate_data(**vars(args), vehicle=vehicle)



import autograd.numpy as np
from autograd import jacobian, hessian
from autograd.numpy import sin, cos, arctan, pi, arctan2, sqrt


class RobotMove:

    dt : float = 1.0

    def __init__(self, state_outlier_flag=False, 
                measurement_outlier_flag=False, noise_type='Gaussian'):
        
        self.dim_x = 2
        self.dim_y = 4
        self.x0 = np.array([0., 0.])
        self.P0 = np.eye(self.dim_x)

        self.state_outlier_flag = state_outlier_flag
        self.measurement_outlier_flag = measurement_outlier_flag
        self.noise_type = noise_type
        self.alpha = 2.0
        self.beta = 5.0

        self.obs_var = np.ones(self.dim_y) * 0.01
        self.Q = np.eye(self.dim_x) * 0.0025
        if noise_type == 'Beta':
            self.R = np.eye(self.dim_y) * (self.alpha * self.beta) / ((self.alpha + self.beta) ** 2 * (self.alpha + self.beta + 1))
        else:
            self.R = np.eye(self.dim_y) * 0.01

    def f(self, x, dt=None):
        if dt is None:
            dt = self.dt
        x1, x2 = x
        x1_ = x1 + dt
        x2_ = x2 + dt
        return np.array([x1_, x2_])

    def h(self, x):
        landmarks = np.array([[-1, 2], [5, 10], [12, 14], [18, 21]])
        hx = []
        px, py = x
        for i in range(len(landmarks)):
            dist = sqrt((px - landmarks[i][0])**2 + (py - landmarks[i][1])**2)
            hx.append(dist)
        return np.array(hx)
        

    def jac_f(self, x, u=0):
        return np.array([[1, 0], [0, 1]])

    def jac_h(self, x):
        landmarks = np.array([[-1, 2], [5, 10], [12, 14], [18, 21]])
        h_list=[]
        for landmark in landmarks:
            h1 = (x[0]-landmark[0])/(sqrt((x[0]-landmark[0])**2+(x[1]-landmark[1])**2))
            h2 = (x[1] - landmark[1]) / (sqrt((x[0] - landmark[0]) ** 2 + (x[1] - landmark[1]) ** 2))
            h_list.append([h1,h2])
        return np.array(h_list)
    
    def f_withnoise(self, x):
        if self.state_outlier_flag:
            prob = np.random.rand()
            if prob <= 0.95:
                cov = self.Q  # 95%概率使用Q
            else:
                cov = 100 * self.Q  # 5%概率使用100Q
        else:
            cov = self.Q
        return self.f(x) + np.random.multivariate_normal(mean=np.zeros(self.dim_x), cov=cov)
    
    def h_withnoise(self, x):
        if self.noise_type == 'Gaussian':
            if self.measurement_outlier_flag:
                prob = np.random.rand()
                if prob <= 0.95:
                    cov = self.R  # 95%概率使用R
                else:
                    cov = 1000 * self.R  # 5%概率使用100R
            else:
                cov = self.R
            return self.h(x) + np.random.multivariate_normal(mean=np.zeros(self.dim_y), cov=cov)
        elif self.noise_type == 'Beta':
            noise = np.random.beta(self.alpha, self.beta, self.dim_y)
            noise = noise - np.mean(noise)
            return self.h(x) + noise
        else:
            return self.h(x) + np.random.laplace(loc=0, scale=self.obs_var, size=(self.dim_y, ))



import numpy as np
from typing import List, Dict
from pathlib import Path

angle_min = -2.09442138672
angle_increment = 0.00581718236208
angles = np.arange(721) * angle_increment + angle_min
der_x_robot = 0.329578
der_l_lm = 0.02
obstacle_info = [[1.052, -2.695], [4.072, -1.752], [6.028, - 3.324]]


def get_from_mappings(mappings: List[Dict], key, default=None):
    return [m.get(key, default) for m in mappings]


def process_lidar(pose: List[float], lidar: List[float], obstacle_info: List[List[float]]):
    if obstacle_info is None:
        return None
    x, y, theta = pose
    pos = np.array([x, y]) + der_x_robot * np.array([np.cos(theta), np.sin(theta)])  # shape: (2, )
    theta = -theta
    rot = np.array([[np.cos(theta), -np.sin(theta)], [np.sin(theta), np.cos(theta)]])
    lidar = np.array(lidar)  # shape: (721, )
    centers = (np.array(obstacle_info) - pos).dot(rot.T)  # shape: (n, 2)
    point_xy = np.stack([lidar * np.cos(angles), lidar * np.sin(angles)], axis=1)  # shape: (721, 2)
    obs_ranges = []
    obs_angles = []
    for center in centers:
        dist = np.linalg.norm(point_xy - center, axis=1)  # shape: (721, )
        # print(dist)
        target = (dist < 1.9)
        # print(np.sum(target))
        if np.sum(target) == 0:
            return None
        range = lidar[target].mean()
        angle = angles[target].mean()
        obs_ranges.append(range)
        obs_angles.append(angle)
    return obs_ranges + obs_angles


def post_process(filepath, obstacle_info, min_len):
    data = np.load(filepath, allow_pickle=True)['arr_0']  # data: list of optional [dict | None]
    processed = []
    temp_list = []
    abnormal = 0
    for d in data:
        if d is None:
            if temp_list:
                processed.append(temp_list)
                temp_list = []
        else:
            d['obs'] = process_lidar(d['state'], d['lidar'], obstacle_info)
            if d['obs'] is None:
                abnormal += 1
                temp_list.clear()
                continue
            temp_list.append(d)
            if len(temp_list) >= min_len:
                processed.append(temp_list)
                temp_list = []
    # filter too short
    processed = [p for p in processed if len(p) >= min_len]
    state = [get_from_mappings(p, 'state') for p in processed]
    action = [get_from_mappings(p, 'control') for p in processed]
    time = [get_from_mappings(p, 'time') for p in processed]
    obs = [get_from_mappings(p, 'obs') for p in processed]
    lidar = [get_from_mappings(p, 'lidar') for p in processed]
    print('num of abnormal frame: ', abnormal)
    return state, action, time, obs, lidar


def load_data(filepath, min_len):
    state, action, time, obs, lidar = post_process(filepath, obstacle_info, min_len=min_len)
    return state, action, obs


if __name__ == '__main__':
    filepath = r'/data_processing\20230216-140452.npz'
    min_len = 100
    state, action, obs = load_data(filepath, min_len)
